function [GCsig prb] = CP_GCTime_MIBCA(X,NLAGS,PVAL)
% -----------------------------------------------------------------------
%   FUNCTION: CP_GCTime.m
%   PURPOSE:  granger causality correlation purged.
%
%   INPUT:  X           -   nvar (rows) by nobs (cols) observation matrix
%           NLAGS       -   number of lags to include in model
%           PVAL        -   p-value (recomended 0.05)
%
%   OUTPUT: GCsig       -   Significant Causality
%           IGC         -   Instantaneous Causality
% -----------------------------------------------------------------------

%Developed by J. Rodrigues
%%%%

if nargin < 3
    PVAL = 0.05;
end
nobs = size(X,2);
nvar = size(X,1);
if(nvar>nobs)
    error('error in CP_GCTime: nvar>nobs, check input matrix');
end

cpgc=tGC_0lag_out(X,NLAGS);

prb = cpgc.prb;
nvar = length(prb);
PR = zeros(nvar);
% Bonferroni correction
q = PVAL./(nvar*(nvar-1));
PR(prb<q) = 1;
%%%%%%%%%%%%%%%%%%%%%%%
GC = cpgc.gc;
% % % GCsig = GC.*PR;
% % % IGC=cpgc.igc;
GCsig=GC;

function [ret] = tGC_0lag_out(X,nlags)
% -----------------------------------------------------------------------
%   FUNCTION: tGC_0lag_out.m
%   PURPOSE:  granger causality correlation purged.
%
%   INPUT:  X           -   nvar (rows) by nobs (cols) observation matrix
%           nlags       -   number of lags to include in model
%
%   OUTPUT: ret.covu    -   covariance of residuals for unrestricted model
%           ret.covr    -   covariance of residuals for restricted models
%           ret.prb     -   Granger causality probabilities (column causes
%                           row. NaN for non-calculated entries)
%           ret.fs      -   F-statistics for above
%           ret.gc      -   log ratio causality magnitude
%           ret.igc     -   instantaneous causality
%           ret.doi     -   difference-of-influence (based on ret.gc)
% -----------------------------------------------------------------------

% figure regression parameters
nobs = size(X,2);
nlagsP1=nlags+1;
nvar = size(X,1);


% remove sample means if present 
m = mean(X');
if(abs(sum(m)) > 0.0001)
    mall = repmat(m',1,nobs);
    X = X-mall;
end

% construct lag matrices
lags = -999*ones(nvar,nobs-nlagsP1,nlagsP1);
for jj=1:nvar
    for ii=1:nlagsP1
        lags(jj,:,nlagsP1-ii+1) = X(jj,ii:nobs-nlagsP1+ii-1);
    end
end

%  regression (no constant term)
regressors = zeros(nobs-nlagsP1,nvar*nlagsP1);
for ii=1:nvar,
    s1 = (ii-1)*nlagsP1+1;
    regressors(:,s1:s1+nlagsP1-1) = squeeze(lags(ii,:,:));
end

new_beta=zeros(nvar*nlagsP1,nvar);
for ii=1:nvar
    xvec = X(ii,:)';
    xdep = xvec(nlags+1:end-1);
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
    temp_regressors=regressors;
    loc=(ii-1)*nlagsP1+1;
    locArray=((1:nvar)-1)*nlagsP1+1;
    temp_regressors(:,loc)=[];
    beta(:,ii) = temp_regressors\xdep;
    new_beta(1:loc-1,ii)=beta(1:loc-1,ii);
    new_beta(loc,ii)=0;
    new_beta(loc+1:end,ii)=beta(loc:end,ii);
    %%%%%%%%%%para prever apenas com os coeficientes passados.
    temp_beta=new_beta(:,ii);
    temp_beta(locArray)=[];
    temp_regressors=regressors;
    temp_regressors(:,locArray)=[];
    %%%%%%%%%%%para prever apenas com os coeficientes passados.
    xpred(:,ii) = temp_regressors*temp_beta;  % temp_beta to have the same number of vars as temp_regressors
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %     beta(:,ii) = regressors\xdep;
    %     xpred(:,ii) = regressors*beta(:,ii);  % keep hold of predicted values
    u(:,ii) = xdep-xpred(:,ii);
    RSS1(ii) = sum(u(:,ii).^2);
    C(ii) = calc_cov(u(:,ii),u(:,ii),nobs-nlagsP1);
end

covu = cov(u);

for ii=1:nvar
    xvec = X(ii,:)';
    xdep = xvec(nlags+1:end-1);          % dependent variable
    caus_inx = setdiff(1:nvar,ii);
    u_r = zeros(nobs-nlagsP1,nvar,'single');
    new_beta=zeros(length(caus_inx)*nlagsP1,1);
    for jj=1:length(caus_inx)
        eq_inx = setdiff(1:nvar,caus_inx(jj));  % vars to include in restricted regression (jj on ii)
        regressors = zeros(nobs-nlagsP1,length(eq_inx)*nlagsP1);
        for kk=1:length(eq_inx)
            s1 = (kk-1)*nlagsP1+1;
            regressors(:,s1:s1+nlagsP1-1) = squeeze(lags(eq_inx(kk),:,:));
        end
        temp_regressors=regressors;
        loc=(jj-1)*nlagsP1+1;
        locArray=((1:nvar-1)-1)*nlagsP1+1;
        temp_regressors(:,loc)=[];
        beta_r = temp_regressors\xdep;
        
        new_beta(1:loc-1)=beta_r(1:loc-1);
        new_beta(loc)=0;
        new_beta(loc+1:end)=beta_r(loc:end);
        %%%%%%%%%%para prever apenas com os coeficientes passados.
        temp_beta=new_beta;
        temp_beta(locArray)=[];
        temp_regressors=regressors;
        temp_regressors(:,locArray)=[];
        %%%%%%%%%%%para prever apenas com os coeficientes passados.
        temp_r = xdep-temp_regressors*temp_beta;
        RSS0(ii,caus_inx(jj)) = sum(temp_r.^2);
        S(ii,caus_inx(jj)) = calc_cov(temp_r,temp_r,nobs-nlagsP1);
        u_r(:,caus_inx(jj)) = temp_r;
    end
    covr{ii} = cov(u_r);
end

% init Granger matrices
gc = ones(nvar).*NaN;
igc = ones(nvar).*NaN;
doi = ones(nvar).*NaN;

% Granger f-tests
prb = ones(nvar).*NaN;
ftest = zeros(nvar);
n2 = (nobs-nlags)-(nvar*nlags);
for ii=1:nvar-1
    for jj=ii+1:nvar
        ftest(ii,jj) = ((RSS0(ii,jj)-RSS1(ii))/nlags)/(RSS1(ii)/n2);    % causality jj->ii
        prb(ii,jj) = 1 - calc_cdff(ftest(ii,jj),nlags,n2);
        ftest(jj,ii) = ((RSS0(jj,ii)-RSS1(jj))/nlags)/(RSS1(jj)/n2);    % causality ii->jj
        prb(jj,ii) = 1 - calc_cdff(ftest(jj,ii),nlags,n2);
        gc(ii,jj) = log(S(ii,jj)/C(ii));
        gc(jj,ii) = log(S(jj,ii)/C(jj));
        doi(ii,jj) = gc(ii,jj) - gc(jj,ii);
        doi(jj,ii) = gc(jj,ii) - gc(ii,jj);
        igc(ii,jj) = log(S(ii,jj)*S(jj,ii)/det(covu));
        igc(jj,ii)=igc(ii,jj);
    end
end

ret.gc = gc;
ret.igc = igc;
ret.fs = ftest;
ret.prb = prb;
ret.covu = covu;
ret.covr = covr;
ret.doi = doi;

function  cov_value=calc_cov(X,Y,n)
x_bar=mean(X);
y_bar=mean(Y);
cov_value=0;
for i=1:n
    cov_value=cov_value+X(i)*Y(i);
end
cov_value=cov_value/n-x_bar*y_bar;

function p = calc_cdff(x,v1,v2)
% Cumulative 'F' distribution
%think of replacing by Matlabs CDF

p = 0;
t = (v1 <= 0 | v2 <= 0 | isnan(x) | isnan(v1) | isnan(v2));
p(t) = NaN;
s = (x==Inf) & ~t;
if any(s)
   p(s) = 1;
   t = t | s;
end

k = find(x > 0 & ~t & isfinite(v1) & isfinite(v2));
if any(k), 
    xx = x(k)./(x(k) + v2(k)./v1(k));
    p(k) = betainc(xx, v1(k)/2, v2(k)/2);
end

if any(~isfinite(v1(:)) | ~isfinite(v2(:)))
   k = find(x > 0 & ~t & isfinite(v1) & ~isfinite(v2) & v2>0);
   if any(k)
      p(k) = chi2cdf(v1(k).*x(k),v1(k));
   end
   k = find(x > 0 & ~t & ~isfinite(v1) & v1>0 & isfinite(v2));
   if any(k)
      p(k) = 1 - chi2cdf(v2(k)./x(k),v2(k));
   end
   k = find(x > 0 & ~t & ~isfinite(v1) & v1>0 & ~isfinite(v2) & v2>0);
   if any(k)
      p(k) = (x(k)>=1);
   end
end
