function [GCsig prb] = MVAR_GCTime_MIBCA(X,NLAGS,PVAL)
% -----------------------------------------------------------------------
%   FUNCTION: MVAR_GCTime.m
%   PURPOSE:  granger causality with Partial Correlation Estimation.
%
%   INPUT:  X           -   nvar (rows) by nobs (cols) observation matrix
%           NLAGS       -   number of lags to include in model
%           PVAL        -   p-value (recomended 0.05)
%
%   OUTPUT: GCsig       -   Significant Causality
%           IGC         -   Instantaneous Causality
% -----------------------------------------------------------------------

%Developed by J. Rodrigues
%%%%

if nargin < 3
    PVAL = 0.05;
end
nobs = size(X,2);
nvar = size(X,1);
if(nvar>nobs)
    error('error in MVAR_GCTime: nvar>nobs, check input matrix');
end

ret = tGC_mvar(X,NLAGS);
prb = ret.prb;
nvar = length(prb);
PR = zeros(nvar);
% Bonferroni correction
q = PVAL./(nvar*(nvar-1));
PR(prb<q) = 1;
%%%%%%%%%%%%%%%%%%%%%%%
GC = ret.gc;
% GCsig = GC.*PR;
% IGC=ret.igc;
GCsig=GC;

function [ret] = tGC_mvar(X,nlags)
% -----------------------------------------------------------------------
%   FUNCTION: tGC_mvar.m
%   PURPOSE:  perform multivariate regression with granger causality statistics
%
%   INPUT:  X           -   nvar (rows) by nobs (cols) observation matrix
%           nlags       -   number of lags to include in model
%           STATFLAG    -   if 1 do F-stats
%
%   OUTPUT: ret.covu    -   covariance of residuals for unrestricted model
%           ret.covr    -   covariance of residuals for restricted models
%           ret.prb     -   Granger causality probabilities (column causes
%                           row. NaN for non-calculated entries)
%           ret.fs      -   F-statistics for above
%           ret.gc      -   log ratio causality magnitude
%           ret.igc     -   instantaneous causality
%           ret.doi     -   difference-of-influence (based on ret.gc)
% -----------------------------------------------------------------------

% figure regression parameters
nobs = size(X,2);
nvar = size(X,1);

% remove sample means if present (no constant terms in this regression)
m = mean(X');
if(abs(sum(m)) > 0.0001)
    mall = repmat(m',1,nobs);
    X = X-mall;
end

[ARF,RCF,PE,DC,u] = mvar(X', nlags);
for ii=1:nvar
    C(ii) = calc_cov(u(:,ii),u(:,ii),nobs-nlags);
    RSS1(ii) = sum(u(:,ii).^2);
end
covu = cov(u);

%   restricted regressions (no constant terms)
for ii=1:nvar
    xvec = X(ii,:)';
    xdep = xvec(nlags+1:end);          % dependent variable
    caus_inx = setdiff(1:nvar,ii);
    u_r = zeros(nobs-nlags,nvar,'single');
    for jj=1:length(caus_inx)
        eq_inx = setdiff(1:nvar,caus_inx(jj));  % vars to include in restricted regression (jj on ii)
        [ARF,RCF,PE,DC,temp_temp_r] = mvar(X(eq_inx,:)', nlags);
        temp_r=temp_temp_r(:,find(eq_inx==ii));
        RSS0(ii,caus_inx(jj)) = sum(temp_r.^2);
        S(ii,caus_inx(jj)) = calc_cov(temp_r,temp_r,nobs-nlags); % dec 08
        u_r(:,caus_inx(jj)) = temp_r;        
    end
    covr{ii} = cov(u_r);
end

% calc Granger values
gc = ones(nvar).*NaN;
igc = ones(nvar).*NaN;
doi = ones(nvar).*NaN;
%   do Granger f-tests if required
prb = ones(nvar).*NaN;
ftest = zeros(nvar);
n2 = (nobs-nlags)-(nvar*nlags);
for ii=1:nvar-1
    for jj=ii+1:nvar
        ftest(ii,jj) = ((RSS0(ii,jj)-RSS1(ii))/nlags)/(RSS1(ii)/n2);    % causality jj->ii
        prb(ii,jj) = 1 - calc_cdff(ftest(ii,jj),nlags,n2);
        ftest(jj,ii) = ((RSS0(jj,ii)-RSS1(jj))/nlags)/(RSS1(jj)/n2);    % causality ii->jj
        prb(jj,ii) = 1 - calc_cdff(ftest(jj,ii),nlags,n2);
        gc(ii,jj) = log(S(ii,jj)/C(ii));
        gc(jj,ii) = log(S(jj,ii)/C(jj));
        doi(ii,jj) = gc(ii,jj) - gc(jj,ii);
        doi(jj,ii) = gc(jj,ii) - gc(ii,jj);
        igc(ii,jj) = log(S(ii,jj)*S(jj,ii)/det(covu));
        igc(jj,ii)=igc(ii,jj);
    end
end

%   organize output structure
ret.gc = gc;
ret.igc = igc;
ret.fs = ftest;
ret.prb = prb;
ret.covu = covu;
ret.covr = covr;
ret.doi = doi;

function  cov_value=calc_cov(X,Y,n)
x_bar=mean(X);
y_bar=mean(Y);
cov_value=0;
for i=1:n
    cov_value=cov_value+X(i)*Y(i);
end
cov_value=cov_value/n-x_bar*y_bar;

function p = calc_cdff(x,v1,v2)
% Cumulative 'F' distribution
%think of replacing by Matlabs CDF

p = 0;
t = (v1 <= 0 | v2 <= 0 | isnan(x) | isnan(v1) | isnan(v2));
p(t) = NaN;
s = (x==Inf) & ~t;
if any(s)
   p(s) = 1;
   t = t | s;
end

k = find(x > 0 & ~t & isfinite(v1) & isfinite(v2));
if any(k), 
    xx = x(k)./(x(k) + v2(k)./v1(k));
    p(k) = betainc(xx, v1(k)/2, v2(k)/2);
end

if any(~isfinite(v1(:)) | ~isfinite(v2(:)))
   k = find(x > 0 & ~t & isfinite(v1) & ~isfinite(v2) & v2>0);
   if any(k)
      p(k) = chi2cdf(v1(k).*x(k),v1(k));
   end
   k = find(x > 0 & ~t & ~isfinite(v1) & v1>0 & isfinite(v2));
   if any(k)
      p(k) = 1 - chi2cdf(v2(k)./x(k),v2(k));
   end
   k = find(x > 0 & ~t & ~isfinite(v1) & v1>0 & ~isfinite(v2) & v2>0);
   if any(k)
      p(k) = (x(k)>=1);
   end
end


function [ARF,RCF,PE,DC,F] = mvar(Y, Pmax)
% MVAR estimates Vieira-Morf Method with unbiased covariance estimation

% Inititialization
[N,M] = size(Y);

if nargin<2,
    Pmax=max([N,M])-1;
end;

if iscell(Y)
    Pmax = min(max(N ,M ),Pmax);
    C    = Y;
end;

[C(:,1:M),n] = covm(Y,'M');
PE(:,1:M)  = C(:,1:M)./n;
%%%%% Partial Correlation Estimation: Vieira-Morf Method with unbiased covariance estimation

%C(:,1:M) = C(:,1:M)/N;
F = Y;
B = Y;
%PEF = C(:,1:M);
%PEB = C(:,1:M);
PEF = PE(:,1:M);
PEB = PE(:,1:M);
for K = 1:Pmax,
    [D,n]	= covm(F(K+1:N,:),B(1:N-K,:),'M');
    D = D./n;
    
    ARF(:,K*M+(1-M:0)) = D / PEB;
    ARB(:,K*M+(1-M:0)) = D'/ PEF;
    
    tmp        = F(K+1:N,:) - B(1:N-K,:)*ARF(:,K*M+(1-M:0)).';
    B(1:N-K,:) = B(1:N-K,:) - F(K+1:N,:)*ARB(:,K*M+(1-M:0)).';
    F(K+1:N,:) = tmp;
    
    for L = 1:K-1,
        tmp      = ARF(:,L*M+(1-M:0))   - ARF(:,K*M+(1-M:0))*ARB(:,(K-L)*M+(1-M:0));
        ARB(:,(K-L)*M+(1-M:0)) = ARB(:,(K-L)*M+(1-M:0)) - ARB(:,K*M+(1-M:0))*ARF(:,L*M+(1-M:0));
        ARF(:,L*M+(1-M:0))   = tmp;
    end;
    
    RCF(:,K*M+(1-M:0)) = ARF(:,K*M+(1-M:0));
    RCB(:,K*M+(1-M:0)) = ARB(:,K*M+(1-M:0));
    
    [PEF,n] = covm(F(K+1:N,:),F(K+1:N,:),'M');
    PEF = PEF./n;
    
    [PEB,n] = covm(B(1:N-K,:),B(1:N-K,:),'M');
    PEB = PEB./n;
    
    PE(:,K*M+(1:M)) = PEF;
end;
F=F(Pmax+1:end,:);

if any(ARF(:)==inf),
    % Test for matrix division bug.
    % This bug was observed in LNX86-ML5.3, 6.1 and 6.5, but not in SGI-ML6.5, LNX86-ML6.5, Octave 2.1.35-40; Other platforms unknown.
    p = 3;
    FLAG_MATRIX_DIVISION_ERROR = ~all(all(isnan(repmat(0,p)/repmat(0,p)))) | ~all(all(isnan(repmat(nan,p)/repmat(nan,p))));
    
    if FLAG_MATRIX_DIVISION_ERROR,
        %fprintf(2,'### Warning MVAR: Bug in Matrix-Division 0/0 and NaN/NaN yields INF instead of NAN.  Workaround is applied.\n');
        warning('MVAR: bug in Matrix-Division 0/0 and NaN/NaN yields INF instead of NAN.  Workaround is applied.');
        
        %%%%% Workaround
        ARF(ARF==inf)=NaN;
        RCF(RCF==inf)=NaN;
    end;
end;

%MAR   = zeros(M,M*Pmax);
DC     = zeros(M);
for K  = 1:Pmax,
    %       VAR{K+1} = -ARF(:,K*M+(1-M:0))';
    DC  = DC + ARF(:,K*M+(1-M:0))'.^2; %DC meausure [3]
end;

function [CC,NN] = covm(X,Y,Mode)
% COVM generates covariance matrix
% X and Y can contain missing values encoded with NaN.

if nargin<3,
        if nargin==2,
		if isnumeric(Y),
                        Mode='M';
                else
		        Mode=Y;
                        Y=[];
                end;
        elseif nargin==1,
                Mode = 'M';
                Y = [];
        elseif nargin==0,
                error('Missing argument(s)');
        end;
end;        

Mode = upper(Mode);

[r1,c1]=size(X);
if ~isempty(Y)
        [r2,c2]=size(Y);
        if r1~=r2,
                error('X and Y must have the same number of observations (rows).');
                return;
        end;
else
        [r2,c2]=size(X);
end;

if (c1>r1) | (c2>r2),
        warning('Covariance is ill-defined, because of too less observations (rows)');
end;

if ~isempty(Y),
        if (~any(Mode=='D') & ~any(Mode=='E')), % if Mode == M
        	NN = real(~isnan(X)')*real(~isnan(Y));
	        X(isnan(X)) = 0; % skip NaN's
	        Y(isnan(Y)) = 0; % skip NaN's
        	CC = X'*Y;
        else  % if any(Mode=='D') | any(Mode=='E'), 
	        [S1,N1] = sumskipnan(X,1);
                [S2,N2] = sumskipnan(Y,1);
                
                NN = real(~isnan(X)')*real(~isnan(Y));
        
	        if any(Mode=='D'), % detrending mode
        		X  = X - ones(r1,1)*(S1./N1);
                        Y  = Y - ones(r1,1)*(S2./N2);
                        if any(Mode=='1'),  %  'D1'
                                NN = NN;
                        else   %  'D0'       
                                NN = max(NN-1,0);
                        end;
                end;
                
                X(isnan(X)) = 0; % skip NaN's
        	Y(isnan(Y)) = 0; % skip NaN's
                CC = X'*Y;
                
                if any(Mode=='E'), % extended mode
                        NN = [r1, N2; N1', NN];
                        CC = [r1, S2; S1', CC];
                end;
	end;
        
else        
        if (~any(Mode=='D') & ~any(Mode=='E')), % if Mode == M
        	tmp = real(~isnan(X));
                NN  = tmp'*tmp; 
                X(isnan(X)) = 0; % skip NaN's
	        CC = X'*X;
        else  % if any(Mode=='D') | any(Mode=='E'), 
	        [S,N] = sumskipnan(X,1);
        	tmp = real(~isnan(X));
                NN  = tmp'*tmp; 
                if any(Mode=='D'), % detrending mode
	                X  = X - ones(r1,1)*(S./N);
                        if any(Mode=='1'),  %  'D1'
                                NN = NN;
                        else  %  'D0'      
                                NN = max(NN-1,0);
                        end;
                end;
                
                X(isnan(X)) = 0; % skip NaN's
                CC = X'*X;
                
                if any(Mode=='E'), % extended mode
                        NN = [r1, N; N', NN];
                        CC = [r1, S; S', CC];
                end;
	end
end;

if nargout<2
        CC = CC./NN; % unbiased
end;
return; 